//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by in_vgm.rc
//
#define MainIcon                        101
#define TabIcons                        102
#define LogoBitmap                      103
#define DlgConfigMain                   201
#define DlgFileInfo                     202
#define DlgCfgPlayback                  203
#define DlgCfgTags                      204
#define DlgCfgVgm7z                     205
#define DlgCfgMuting                    206
#define DlgCfgOptPan                    207
#define TabCollection                   1001
#define ImmediateUpdCheck               1002
#define HelpButton                      1003
#define LoopText                        1101
#define FadeText                        1102
#define PauseNlText                     1103
#define PauseLpText                     1104
#define RateRecRadio                    1105
#define Rate60HzRadio                   1106
#define Rate50HzRadio                   1107
#define RateOtherRadio                  1108
#define RateText                        1109
#define RateLabel                       1110
#define VolumeSlider                    1111
#define VolumeText                      1112
#define SmplPbRateText                  1113
#define ResmpModeList                   1114
#define SmplChipRateText                1115
#define ChipSmpModeList                 1116
#define SurroundCheck                   1117
#define SmplChipRateLabel               1118
#define SmplChipRateHzLabel             1119
#define LoopTimesLabel                  1120
#define TitleFormatText                 1201
#define PreferJapCheck                  1202
#define FM2413Check                     1203
#define TrimWhitespcCheck               1204
#define StdSeparatorsCheck              1205
#define MLTypeText                      1206
#define TagFallbackCheck                1207
#define NoInfoCacheCheck                1208
#define MutingChipList                  1301
#define MutingChipNumList               1302
#define MuteChn1Check                   1303
#define MuteChn2Check                   1304
#define MuteChn3Check                   1305
#define MuteChn4Check                   1306
#define MuteChn5Check                   1307
#define MuteChn6Check                   1308
#define MuteChn7Check                   1309
#define MuteChn8Check                   1310
#define MuteChn9Check                   1311
#define MuteChn10Check                  1312
#define MuteChn11Check                  1313
#define MuteChn12Check                  1314
#define MuteChn13Check                  1315
#define MuteChn14Check                  1316
#define MuteChn15Check                  1317
#define MuteChn16Check                  1318
#define MuteChn17Check                  1319
#define MuteChn18Check                  1320
#define MuteChn19Check                  1321
#define MuteChn20Check                  1322
#define MuteChn21Check                  1323
#define MuteChn22Check                  1324
#define MuteChn23Check                  1325
#define MuteChn24Check                  1326
#define ResetMuteCheck                  1327
#define MuteChipCheck                   1328
#define TrkTitleText                    2001
#define TrkAuthorText                   2002
#define GameNameText                    2003
#define GameSysText                     2004
#define GameDateText                    2005
#define VGMFileText                     2006
#define VGMCrtText                      2007
#define VGMVerText                      2008
#define VGMGainText                     2009
#define VGMSizeText                     2010
#define VGMLenText                      2011
#define VGMNotesText                    2012
#define ChipUseText                     2013
#define LangEngCheck                    2014
#define LangJapCheck                    2015
#define ConfigPluginButton              2016
#define BrwsrInfoButton                 2017
#define EmuOptChipList                  3001
#define EmuOptChipNumList               3002
#define EmuCoreRadio1                   3003
#define EmuCoreRadio2                   3004
#define PanChn1Label                    3005
#define PanChn2Label                    3006
#define PanChn3Label                    3007
#define PanChn4Label                    3008
#define PanChn5Label                    3009
#define PanChn6Label                    3010
#define PanChn7Label                    3011
#define PanChn8Label                    3012
#define PanChn9Label                    3013
#define PanChn10Label                   3014
#define PanChn11Label                   3015
#define PanChn12Label                   3016
#define PanChn13Label                   3017
#define PanChn14Label                   3018
#define PanChn15Label                   3019
#define PanChn1Slider                   3020
#define PanChn2Slider                   3021
#define PanChn3Slider                   3022
#define PanChn4Slider                   3023
#define PanChn5Slider                   3024
#define PanChn6Slider                   3025
#define PanChn7Slider                   3026
#define PanChn8Slider                   3027
#define PanChn9Slider                   3028
#define PanChn10Slider                  3029
#define PanChn11Slider                  3030
#define PanChn12Slider                  3031
#define PanChn13Slider                  3032
#define PanChn14Slider                  3033
#define PanChn15Slider                  3034

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        212
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         3038
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
